from __future__ import annotations

import contextlib
import dataclasses
import os
import pathlib
import sys
import types
from collections.abc import Iterator, Mapping, Sequence
from typing import TYPE_CHECKING, Generic, cast

from django.apps.registry import Apps
from django.conf import LazySettings
from typing_extensions import Self

from . import protocols


@dataclasses.dataclass(frozen=True, kw_only=True)
class replaced_env_vars_and_sys_path:
    """
    Helper to modify sys.path and os.environ such that those changes are reversed
    upon exiting the contextmanager
    """

    additional_sys_path: Sequence[str]
    env_vars: Mapping[str, str]

    undo_env: dict[str, str | None] = dataclasses.field(init=False, default_factory=dict)
    remove_path: list[str] = dataclasses.field(init=False, default_factory=list)

    def __enter__(self) -> None:
        # Determine what to undo later
        for k, v in self.env_vars.items():
            if k not in os.environ:
                self.undo_env[k] = None
            else:
                self.undo_env[k] = os.environ[k]

        for path in self.additional_sys_path:
            if path not in sys.path:
                self.remove_path.append(path)

        # Make the change itself
        for path in self.additional_sys_path:
            if path not in sys.path:
                sys.path.append(path)

        for k, v in self.env_vars.items():
            os.environ[k] = v

    def __exit__(self, exc_type: type[Exception], tb: types.TracebackType, exc: Exception) -> None:
        for path in self.remove_path:
            if path in sys.path:
                sys.path.remove(path)

        for k, v in self.undo_env.items():
            if v is None:
                if k in os.environ:
                    del os.environ[k]
            else:
                os.environ[k] = v


@dataclasses.dataclass(frozen=True, kw_only=True)
class Project:
    root_dir: pathlib.Path
    additional_sys_path: Sequence[str]
    env_vars: Mapping[str, str]

    discovery: protocols.Discovery[Self]

    @contextlib.contextmanager
    def setup_sys_path_and_env_vars(self) -> Iterator[None]:
        with replaced_env_vars_and_sys_path(
            additional_sys_path=self.additional_sys_path, env_vars=self.env_vars
        ):
            yield

    @contextlib.contextmanager
    def instantiate_django(self) -> Iterator[protocols.Loaded[Self]]:
        with self.setup_sys_path_and_env_vars():
            from django.apps import apps
            from django.conf import settings

            if not settings.configured:
                settings._setup()  # type: ignore[misc]
            apps.populate(settings.INSTALLED_APPS)

            assert apps.apps_ready, "Apps are not ready"
            assert settings.configured, "Settings are not configured"

            yield Loaded(
                project=self,
                root_dir=self.root_dir,
                env_vars=self.env_vars,
                settings=settings,
                apps=apps,
                discovery=self.discovery,
            )

    def load_project(self) -> protocols.Loaded[Self]:
        with self.instantiate_django() as loaded_project:
            return loaded_project


@dataclasses.dataclass(frozen=True, kw_only=True)
class Loaded(Generic[protocols.T_Project]):
    project: protocols.T_Project
    root_dir: pathlib.Path
    env_vars: Mapping[str, str]
    settings: LazySettings
    apps: Apps

    discovery: protocols.Discovery[protocols.T_Project]

    def perform_discovery(self) -> protocols.Discovered[protocols.T_Project]:
        installed_models_modules = self.discovery.discover_installed_models(self)

        all_models: dict[protocols.ImportPath, protocols.Model] = {}
        for module in installed_models_modules.values():
            all_models.update(module.defined_models)

        return Discovered(
            loaded_project=self,
            all_models=all_models,
            installed_apps=self.settings.INSTALLED_APPS,
            settings_types=self.discovery.discover_settings_types(self),
            installed_models_modules=installed_models_modules,
            concrete_models=self.discovery.discover_concrete_models(self, all_models),
        )


@dataclasses.dataclass(frozen=True, kw_only=True)
class Discovered(Generic[protocols.T_Project]):
    loaded_project: Loaded[protocols.T_Project]

    all_models: protocols.ModelMap
    installed_apps: list[str]
    settings_types: protocols.SettingsTypesMap
    concrete_models: protocols.ConcreteModelsMap
    installed_models_modules: protocols.ModelModulesMap


if TYPE_CHECKING:
    _P: protocols.Project = cast(Project, None)
    _LP: protocols.P_Loaded = cast(Loaded[protocols.P_Project], None)
    _AP: protocols.P_Discovered = cast(Discovered[protocols.P_Project], None)

    C_Project = Project
    _CLP: protocols.Loaded[C_Project] = cast(Loaded[C_Project], None)
    _CAP: protocols.Discovered[C_Project] = cast(Discovered[C_Project], None)



import types
from unittest import mock

from mypy.options import Options
from mypy.plugin import AttributeContext, Plugin
from mypy.types import AnyType, TypeOfAny
from mypy.types import Type as MypyType

from extended_mypy_django_plugin.plugin import hook


class TestHookAndDecorator:
    def test_it_caches_instances_of_the_hook_class(self) -> None:
        result = AnyType(TypeOfAny.from_error)
        called: list[object] = []

        class MyPlugin(Plugin):
            @hook.hook
            class get_attribute_hook(hook.Hook["MyPlugin", AttributeContext, MypyType]):
                def hook(self, fullname: str) -> hook.MypyHook[AttributeContext, MypyType]:
                    called.append(
                        ("hook", self, fullname, self.plugin, self.super_hook_maker(fullname))
                    )
                    return self.run

                def run(self, ctx: AttributeContext) -> MypyType:
                    called.append(("run", self, ctx))
                    return result

        options = Options()
        plugin = MyPlugin(options)
        assert called == []

        descriptor = MyPlugin.get_attribute_hook
        assert isinstance(descriptor, hook.hook)

        assert descriptor._hook_instance is None
        hk = plugin.get_attribute_hook("one")
        assert isinstance(hk, types.MethodType)

        hook_instance = MyPlugin.get_attribute_hook._hook_instance
        assert hk.__self__ is hook_instance
        assert hk == hook_instance.run

        assert called == [("hook", hook_instance, "one", plugin, None)]

        called.clear()
        ctx = mock.Mock(spec=AttributeContext)
        assert hk(ctx) is result
        assert called == [("run", hook_instance, ctx)]

        # And when we call it again, we use the same hook instance
        called.clear()
        hk2 = plugin.get_attribute_hook("two")
        assert isinstance(hk2, types.MethodType)

        assert MyPlugin.get_attribute_hook._hook_instance is hook_instance
        assert hk2.__self__ is hook_instance
        assert hk2 == hook_instance.run

        assert called == [("hook", hook_instance, "two", plugin, None)]

        called.clear()
        ctx = mock.Mock(spec=AttributeContext)
        assert hk2(ctx) is result
        assert called == [("run", hook_instance, ctx)]



def interface____differentiated__1() -> None:
    return None

mod = "django.contrib.admin.models"
summary = "__virtual__.mod_2456226428::django.contrib.admin.models::installed_apps=__installed_apps_hash__::significant=3233719984"

import django.contrib.admin.models
import django.db.models
ConcreteQuerySet__LogEntry = django.db.models.QuerySet[django.contrib.admin.models.LogEntry]
Concrete__LogEntry = django.contrib.admin.models.LogEntry



import dataclasses
from typing import TYPE_CHECKING, Generic, cast

from .. import protocols
from . import concrete_models, known_models, settings_types


@dataclasses.dataclass
class Discovery(Generic[protocols.T_Project]):
    discover_settings_types: protocols.SettingsTypesDiscovery[protocols.T_Project] = (
        dataclasses.field(default_factory=settings_types.NaiveSettingsTypesDiscovery)
    )
    discover_installed_models: protocols.InstalledModelsDiscovery[protocols.T_Project] = (
        dataclasses.field(default_factory=known_models.DefaultInstalledModulesDiscovery)
    )
    discover_concrete_models: protocols.ConcreteModelsDiscovery[protocols.T_Project] = (
        dataclasses.field(default_factory=concrete_models.ConcreteModelsDiscovery)
    )


if TYPE_CHECKING:
    _A: protocols.P_Discovery = cast(Discovery[protocols.P_Project], None)



from __future__ import annotations

import dataclasses
from collections.abc import Sequence
from typing import TYPE_CHECKING, Protocol, cast

from django.db import models
from typing_extensions import Self

from extended_mypy_django_plugin.django_analysis import ImportPath, protocols


class FieldCreator(Protocol):
    def __call__(
        self, *, model_import_path: protocols.ImportPath, field: protocols.DjangoField
    ) -> protocols.Field: ...


def _find_default_custom_queryset(model: type[models.Model]) -> protocols.ImportPath | None:
    """
    Given a model, find the import path to the custom default queryset associated with that model
    if it has one.
    """
    if (default_manager := model._meta.default_manager) is None:
        # No custom default queryset if there is no default manager
        return None

    if (qs := getattr(default_manager, "_queryset_class", models.QuerySet)) is models.QuerySet:
        # default queryset isn't interesting to us!
        return None

    if not isinstance(qs, type):
        # If the queryset isn't a type, then it's unlikely we can find an import path to it
        return None

    return ImportPath.from_cls(qs)


@dataclasses.dataclass(frozen=True, kw_only=True)
class Model:
    @classmethod
    def create(
        cls,
        *,
        field_creator: FieldCreator,
        model: type[models.Model],
    ) -> Self:
        return cls(
            model_name=model.__qualname__,
            module_import_path=ImportPath.cls_module(model),
            import_path=(model_import_path := ImportPath.from_cls(model)),
            is_abstract=model._meta.abstract,
            default_custom_queryset=_find_default_custom_queryset(model),
            all_fields={
                # We want to know the concrete set of models, and we include hidden fields
                # So that we can see all related models later on
                field.name: field_creator(model_import_path=model_import_path, field=field)
                for field in model._meta.get_fields(include_parents=True, include_hidden=True)
            },
            models_in_mro=[
                # Only care about parent models that are themselves other models and aren't
                # the base django Model class or the model being looked at
                ImportPath.from_cls(m)
                for m in model.__mro__
                if issubclass(m, models.Model) and m not in (models.Model, model)
            ],
        )

    model_name: str
    module_import_path: protocols.ImportPath
    import_path: protocols.ImportPath
    is_abstract: bool
    default_custom_queryset: protocols.ImportPath | None
    all_fields: protocols.FieldsMap
    models_in_mro: Sequence[protocols.ImportPath]


if TYPE_CHECKING:
    _M: protocols.Model = cast(Model, None)



import dataclasses
import types
from collections.abc import Sequence
from typing import TYPE_CHECKING, Protocol, cast

from django.db import models
from typing_extensions import Self

from extended_mypy_django_plugin.django_analysis import ImportPath, protocols


class ModelCreator(Protocol):
    def __call__(self, *, model: type[models.Model]) -> protocols.Model: ...


@dataclasses.dataclass(frozen=True, kw_only=True)
class Module:
    @classmethod
    def create(
        cls,
        *,
        model_creator: ModelCreator,
        import_path: protocols.ImportPath,
        module: types.ModuleType,
        models: Sequence[type[models.Model]],
    ) -> Self:
        return cls(
            import_path=import_path,
            defined_models={
                ImportPath.from_cls(model): model_creator(model=model) for model in models
            },
        )

    import_path: protocols.ImportPath
    defined_models: protocols.ModelMap


if TYPE_CHECKING:
    _M: protocols.Module = cast(Module, None)



import pathlib
import textwrap

from extended_mypy_django_plugin_test_driver import ScenarioBuilder


class TestRenamingQuerySet:
    def test_can_move_queryset_from_abstract_class_to_concrete_children(
        self, builder: ScenarioBuilder, tmp_path: pathlib.Path
    ) -> None:
        builder.set_installed_apps("parent", "child1", "child2")
        for app in ("parent", "child1", "child2"):
            builder.on(f"{app}/__init__.py").set("")
            builder.on(f"{app}/apps.py").set(
                f"""
                from django.apps import AppConfig

                class Config(AppConfig):
                    name = "{app}"
                """,
            )

        builder.on("parent/models.py").set(
            """
            from typing import TYPE_CHECKING, Union
            from django.db import models


            if TYPE_CHECKING:
                from child1.models import Child1
                from child2.models import Child2

            class ParentQuerySet(models.QuerySet[Union["Child1", "Child2"]]):
                pass

            ParentManager = models.Manager.from_queryset(ParentQuerySet)


            class Parent(models.Model):
                objects = ParentManager()

                class Meta:
                    abstract = True
            """
        )
        builder.on("child1/models.py").set(
            """
            from parent.models import Parent

            class Child1(Parent):
                pass
            """
        )
        builder.on("child2/models.py").set(
            """
            from parent.models import Parent

            class Child2(Parent):
                pass
            """
        )

        deps_dest = tmp_path / ".mypy_django_deps"

        builder.populate_virtual_deps(deps_dest=deps_dest)

        expected = {
            "mod_3961720227.py": """
                def interface__timestamp() -> None:
                    return None

                mod = "django.contrib.contenttypes.models"
                summary = "__virtual_extended_mypy_django_plugin_report__.mod_3961720227::django.contrib.contenttypes.models::installed_apps=3376484868::significant=438215680"

                import django.contrib.contenttypes.models
                import django.db.models
                ConcreteQuerySet__ContentType = django.db.models.QuerySet[django.contrib.contenttypes.models.ContentType]
                Concrete__ContentType = django.contrib.contenttypes.models.ContentType
                """,
            "mod_566232296.py": """
                def interface__timestamp() -> None:
                    return None

                mod = "child1.models"
                summary = "__virtual_extended_mypy_django_plugin_report__.mod_566232296::child1.models::installed_apps=3376484868::significant=967422822"

                import child1.models
                import parent.models
                ConcreteQuerySet__Child1 = parent.models.ParentQuerySet
                Concrete__Child1 = child1.models.Child1
                """,
            "mod_566756585.py": """
                def interface__timestamp() -> None:
                    return None

                mod = "child2.models"
                summary = "__virtual_extended_mypy_django_plugin_report__.mod_566756585::child2.models::installed_apps=3376484868::significant=1344320379"

                import child2.models
                import parent.models
                ConcreteQuerySet__Child2 = parent.models.ParentQuerySet
                Concrete__Child2 = child2.models.Child2
                """,
            "mod_614729021.py": """
                def interface__timestamp() -> None:
                    return None

                mod = "parent.models"
                summary = "__virtual_extended_mypy_django_plugin_report__.mod_614729021::parent.models::installed_apps=3376484868::significant=309881802"

                import child1.models
                import child2.models
                import parent.models
                ConcreteQuerySet__Parent = parent.models.ParentQuerySet | parent.models.ParentQuerySet
                Concrete__Parent = child1.models.Child1 | child2.models.Child2
                """,
        }

        for fle in (deps_dest / "__virtual_extended_mypy_django_plugin_report__").iterdir():
            want = textwrap.dedent(expected[fle.name]).strip()
            assert fle.read_text().strip() == want

        builder.on("parent/models.py").set(
            """
            from typing import TYPE_CHECKING, Union, TypeVar
            from extended_mypy_django_plugin import Concrete
            from django.db import models

            T_Child = TypeVar("T_Child", bound=Concrete["Parent"])

            class ParentQuerySet(models.QuerySet[T_Child]):
                pass


            class Parent(models.Model):
                class Meta:
                    abstract = True
            """
        )
        builder.on("child1/models.py").set(
            """
            from parent.models import Parent, ParentQuerySet
            from django.db import models

            class Child1QuerySet(ParentQuerySet["Child1"]):
                pass

            Child1Manager = models.Manager.from_queryset(Child1QuerySet)

            class Child1(Parent):
                objects = Child1Manager()
            """
        )
        builder.on("child2/models.py").set(
            """
            from parent.models import Parent, ParentQuerySet
            from django.db import models

            class Child2QuerySet(ParentQuerySet["Child1"]):
                pass

            Child2Manager = models.Manager.from_queryset(Child2QuerySet)


            class Child2(Parent):
                objects = Child2Manager()
            """
        )

        expected["mod_566232296.py"] = """
            def interface__timestamp() -> None:
                return None

            mod = "child1.models"
            summary = "__virtual_extended_mypy_django_plugin_report__.mod_566232296::child1.models::installed_apps=3376484868::significant=2468262588"

            import child1.models
            ConcreteQuerySet__Child1 = child1.models.Child1QuerySet
            Concrete__Child1 = child1.models.Child1
            """

        expected["mod_566756585.py"] = """
            def interface__timestamp() -> None:
                return None

            mod = "child2.models"
            summary = "__virtual_extended_mypy_django_plugin_report__.mod_566756585::child2.models::installed_apps=3376484868::significant=2877928147"

            import child2.models
            ConcreteQuerySet__Child2 = child2.models.Child2QuerySet
            Concrete__Child2 = child2.models.Child2
            """

        expected["mod_614729021.py"] = """
            def interface__timestamp() -> None:
                return None

            mod = "parent.models"
            summary = "__virtual_extended_mypy_django_plugin_report__.mod_614729021::parent.models::installed_apps=3376484868::significant=3300935593"

            import child1.models
            import child2.models
            import parent.models
            ConcreteQuerySet__Parent = child1.models.Child1QuerySet | child2.models.Child2QuerySet
            Concrete__Parent = child1.models.Child1 | child2.models.Child2
            """

        builder.populate_virtual_deps(deps_dest=deps_dest)

        for fle in sorted(
            (deps_dest / "__virtual_extended_mypy_django_plugin_report__").iterdir()
        ):
            want = textwrap.dedent(expected[fle.name]).strip()
            assert fle.read_text().strip() == want

    def test_can_rename_duplicate_concrete_querysets(
        self, builder: ScenarioBuilder, tmp_path: pathlib.Path
    ) -> None:
        builder.set_installed_apps("parent", "child1", "child2")
        for app in ("parent", "child1", "child2"):
            builder.on(f"{app}/__init__.py").set("")
            builder.on(f"{app}/apps.py").set(
                f"""
                from django.apps import AppConfig

                class Config(AppConfig):
                    name = "{app}"
                """,
            )

        builder.on("parent/models.py").set(
            """
            from typing import TYPE_CHECKING, Union
            from django.db import models


            if TYPE_CHECKING:
                from child1.models import Child1
                from child2.models import Child2

            class Parent(models.Model):
                class Meta:
                    abstract = True
            """
        )
        builder.on("child1/models.py").set(
            """
            from parent.models import Parent
            from django.db import models

            class ChildQuerySet(models.QuerySet["Child"]):
                pass

            ChildManager = models.Manager.from_queryset(ChildQuerySet)

            class Child(Parent):
                objects = ChildManager()
            """
        )
        builder.on("child2/models.py").set(
            """
            from parent.models import Parent
            from django.db import models

            class ChildQuerySet(models.QuerySet["Child"]):
                pass

            ChildManager = models.Manager.from_queryset(ChildQuerySet)

            class Child(Parent):
                objects = ChildManager()
            """
        )

        deps_dest = tmp_path / ".mypy_django_deps"

        pathlib.Path("/tmp/debug").unlink(missing_ok=True)
        builder.populate_virtual_deps(deps_dest=deps_dest)

        expected = {
            "mod_3961720227.py": """
                def interface__timestamp() -> None:
                    return None

                mod = "django.contrib.contenttypes.models"
                summary = "__virtual_extended_mypy_django_plugin_report__.mod_3961720227::django.contrib.contenttypes.models::installed_apps=3376484868::significant=438215680"

                import django.contrib.contenttypes.models
                import django.db.models
                ConcreteQuerySet__ContentType = django.db.models.QuerySet[django.contrib.contenttypes.models.ContentType]
                Concrete__ContentType = django.contrib.contenttypes.models.ContentType
                """,
            "mod_566232296.py": """
                def interface__timestamp() -> None:
                    return None

                mod = "child1.models"
                summary = "__virtual_extended_mypy_django_plugin_report__.mod_566232296::child1.models::installed_apps=3376484868::significant=1784590644"

                import child1.models
                ConcreteQuerySet__Child = child1.models.ChildQuerySet
                Concrete__Child = child1.models.Child
                """,
            "mod_566756585.py": """
                def interface__timestamp() -> None:
                    return None

                mod = "child2.models"
                summary = "__virtual_extended_mypy_django_plugin_report__.mod_566756585::child2.models::installed_apps=3376484868::significant=2056565059"

                import child2.models
                ConcreteQuerySet__Child = child2.models.ChildQuerySet
                Concrete__Child = child2.models.Child
                """,
            "mod_614729021.py": """
                def interface__timestamp() -> None:
                    return None

                mod = "parent.models"
                summary = "__virtual_extended_mypy_django_plugin_report__.mod_614729021::parent.models::installed_apps=3376484868::significant=1191855927"

                import child1.models
                import child2.models
                import parent.models
                ConcreteQuerySet__Parent = child1.models.ChildQuerySet | child2.models.ChildQuerySet
                Concrete__Parent = child1.models.Child | child2.models.Child
                """,
        }

        for fle in (deps_dest / "__virtual_extended_mypy_django_plugin_report__").iterdir():
            want = textwrap.dedent(expected[fle.name]).strip()
            assert fle.read_text().strip() == want

        builder.on("child1/models.py").set(
            """
            from parent.models import Parent
            from django.db import models

            class _Child1QuerySet(models.QuerySet["Child"]):
                pass

            ChildQuerySet = _Child1QuerySet
            ChildManager = models.Manager.from_queryset(_Child1QuerySet)

            class Child(Parent):
                objects = ChildManager()
            """
        )
        builder.on("child2/models.py").set(
            """
            from parent.models import Parent
            from django.db import models

            class _Child2QuerySet(models.QuerySet["Child"]):
                pass

            ChildQuerySet = _Child2QuerySet
            ChildManager = models.Manager.from_queryset(_Child2QuerySet)

            class Child(Parent):
                objects = ChildManager()
            """
        )

        expected["mod_566232296.py"] = """
            def interface__timestamp() -> None:
                return None

            mod = "child1.models"
            summary = "__virtual_extended_mypy_django_plugin_report__.mod_566232296::child1.models::installed_apps=3376484868::significant=2733879748"

            import child1.models
            ConcreteQuerySet__Child = child1.models._Child1QuerySet
            Concrete__Child = child1.models.Child
            """

        expected["mod_566756585.py"] = """
            def interface__timestamp() -> None:
                return None

            mod = "child2.models"
            summary = "__virtual_extended_mypy_django_plugin_report__.mod_566756585::child2.models::installed_apps=3376484868::significant=3022762452"

            import child2.models
            ConcreteQuerySet__Child = child2.models._Child2QuerySet
            Concrete__Child = child2.models.Child
            """

        expected["mod_614729021.py"] = """
            def interface__timestamp() -> None:
                return None

            mod = "parent.models"
            summary = "__virtual_extended_mypy_django_plugin_report__.mod_614729021::parent.models::installed_apps=3376484868::significant=1197557559"

            import child1.models
            import child2.models
            import parent.models
            ConcreteQuerySet__Parent = child1.models._Child1QuerySet | child2.models._Child2QuerySet
            Concrete__Parent = child1.models.Child | child2.models.Child
            """

        pathlib.Path("/tmp/debug").write_text("")
        builder.populate_virtual_deps(deps_dest=deps_dest)

        for fle in sorted(
            (deps_dest / "__virtual_extended_mypy_django_plugin_report__").iterdir()
        ):
            want = textwrap.dedent(expected[fle.name]).strip()
            assert fle.read_text().strip() == want



import enum
from collections.abc import Iterator, Mapping, MutableMapping, Sequence, Set
from typing import TYPE_CHECKING, Optional, Protocol, TypeVar

from mypy import errorcodes
from mypy.nodes import SymbolTableNode, TypeInfo
from mypy.plugin import (
    AnalyzeTypeContext,
    AttributeContext,
    DynamicClassDefContext,
    FunctionContext,
    MethodContext,
)
from mypy.types import (
    AnyType,
    Instance,
    PlaceholderType,
    ProperType,
    TypeType,
    TypeVarType,
    UnionType,
)
from mypy.types import Type as MypyType

from ..django_analysis import protocols as d_protocols

T_Report = TypeVar("T_Report", bound="Report")


CombinedReport = d_protocols.CombinedReport
VirtualDependencyHandler = d_protocols.VirtualDependencyHandler

TypeVarMap = MutableMapping[TypeVarType, Instance | TypeType | UnionType]


class KnownClasses(enum.Enum):
    CONCRETE = "extended_mypy_django_plugin.annotations.Concrete"


class KnownAnnotations(enum.Enum):
    CONCRETE = "extended_mypy_django_plugin.annotations.Concrete"
    DEFAULT_QUERYSET = "extended_mypy_django_plugin.annotations.DefaultQuerySet"

    @classmethod
    def resolve(cls, fullname: str) -> Optional["KnownAnnotations"]:
        """
        This function is an alternative to a try..catch(ValueError) that is quicker
        """
        if fullname == cls.CONCRETE.value:
            return cls.CONCRETE

        elif fullname == cls.DEFAULT_QUERYSET.value:
            return cls.DEFAULT_QUERYSET

        else:
            return None


class Report(Protocol):
    def additional_deps(
        self,
        *,
        file_import_path: str,
        imports: Set[str],
        super_deps: Sequence[tuple[int, str, int]],
        django_settings_module: str,
        using_incremental_cache: bool,
    ) -> Sequence[tuple[int, str, int]]:
        """
        This is used to add to the result for the get_additional_deps mypy hook.

        It takes the import path for the file being looked at, any additional deps that have already
        been determined for the file, the imports the file contains as a list of full imports,
        and the import path of the django settings module.

        It must return the full set of additional deps the mypy plugin should use for this file
        """

    def get_concrete_aliases(self, *models: str) -> Mapping[str, str | None]:
        """
        Given import paths to some models, return a map of those models to a type alias
        with the concrete models for that model

        If concrete models cannot be found for a model it's entry will be given as None
        """

    def get_queryset_aliases(self, *models: str) -> Mapping[str, str | None]:
        """
        Given import paths to some models, return a map of those models to a type alias
        with the concrete querysets for that model

        If concrete querysets cannot be found for a model it's entry will be given as None
        """


class FailFunc(Protocol):
    """
    Used to insert an error into the mypy output
    """

    def __call__(self, msg: str, code: errorcodes.ErrorCode | None = None) -> None: ...


class DeferFunc(Protocol):
    """
    Used to tell mypy to defer and come back later

    Returns True if was able to defer
    """

    def __call__(self) -> bool: ...


class LookupInfo(Protocol):
    """
    Given some fullname return a TypeInfo if one can be found
    """

    def __call__(self, fullname: str) -> TypeInfo | None: ...


class NamedTypeOrNone(Protocol):
    """
    Given some fullname and arguments, find the type info for that fullname if can be found and
    return an instance representing that object with those arguments
    """

    def __call__(self, fullname: str, args: list[MypyType] | None = None) -> Instance | None: ...


class AliasGetter(Protocol):
    """
    Given fullnames to zero or more models return a Mapping of those models to type aliases
    for the concrete aliases of that model where the value in the mapping is None if the alias
    could not be found.
    """

    def __call__(self, *models: str) -> Mapping[str, str | None]: ...


class LookupAlias(Protocol):
    """
    Given an alias for the concrete of some model, return Instance of the models represented
    by that type alias
    """

    def __call__(self, alias: str) -> Iterator[Instance | PlaceholderType]: ...


class LookupFullyQualified(Protocol):
    """
    Find a symbol for the provided fullname
    """

    def __call__(self, fullname: str) -> SymbolTableNode | None: ...


class ResolveManagerMethodFromInstance(Protocol):
    """
    Used to fold the fix from https://github.com/typeddjango/django-stubs/pull/2027 into the plugin
    """

    def __call__(
        self, instance: Instance, method_name: str, ctx: AttributeContext
    ) -> MypyType: ...


class Resolver(Protocol):
    """
    Used to resolve concrete annotations
    """

    def resolve(
        self, annotation: KnownAnnotations, model_type: ProperType
    ) -> Instance | TypeType | UnionType | AnyType | PlaceholderType | None:
        """
        Given a specific annotation and some model return the resolved
        concrete form.
        """


ValidContextForAnnotationResolver = (
    DynamicClassDefContext
    | AnalyzeTypeContext
    | AttributeContext
    | MethodContext
    | FunctionContext
)


class ResolverMaker(Protocol):
    """
    This is used to create an instance of Resolver
    """

    def __call__(cls, *, ctx: ValidContextForAnnotationResolver) -> Resolver: ...


class SignatureInfo(Protocol):
    """
    This is used by the type checker to represent important information about a signature for
    a method or function
    """

    @property
    def is_guard(self) -> bool:
        """
        Whether this signature represents a type guard
        """

    @property
    def transformed_ret_type(self) -> ProperType:
        """
        The return type of the signature after resolving annotations
        """


if TYPE_CHECKING:
    P_Report = Report
    P_Resolver = Resolver
    P_ResolverMaker = ResolverMaker
    P_VirtualDependencyHandler = VirtualDependencyHandler[P_Report]



import dataclasses
from typing import TYPE_CHECKING, Generic, cast

from .. import protocols


@dataclasses.dataclass(frozen=True, kw_only=True)
class NaiveSettingsTypesDiscovery(Generic[protocols.T_Project]):
    """
    The default implementation is a little naive and is only able to rely on inspecting
    the values on the settings object.
    """

    def __call__(
        self, loaded_project: protocols.Loaded[protocols.T_Project], /
    ) -> protocols.SettingsTypesMap:
        result: dict[str, str] = {}
        settings = loaded_project.settings

        for name in dir(settings):
            if not self.valid_setting_name(loaded_project=loaded_project, name=name):
                continue

            result[name] = self.type_from_setting(
                loaded_project=loaded_project, name=name, value=getattr(settings, name)
            )

        return result

    def valid_setting_name(
        self, *, loaded_project: protocols.Loaded[protocols.T_Project], name: str
    ) -> bool:
        return not name.startswith("_") and name.isupper()

    def type_from_setting(
        self, *, loaded_project: protocols.Loaded[protocols.T_Project], name: str, value: object
    ) -> str:
        return str(type(value))


if TYPE_CHECKING:
    _STA: protocols.P_SettingsTypesDiscovery = cast(
        NaiveSettingsTypesDiscovery[protocols.P_Project], None
    )



import functools

import pytest
from pytest_typing_runner import parse, protocols

from .scenario import Scenario, ScenarioBuilder, ScenarioFile, ScenarioRunner


@pytest.fixture
def typing_scenario_maker() -> protocols.ScenarioMaker[Scenario]:
    return Scenario.create


@pytest.fixture
def typing_scenario_runner_maker() -> protocols.ScenarioRunnerMaker[Scenario]:
    return ScenarioRunner.create


@pytest.fixture
def builder(typing_scenario_runner: ScenarioRunner) -> ScenarioBuilder:
    return ScenarioBuilder(
        scenario_runner=typing_scenario_runner,
        scenario_file_maker=functools.partial(
            ScenarioFile,
            file_parser=parse.FileContent().parse,
            file_modification=typing_scenario_runner.file_modification,
        ),
    )


@pytest.fixture
def scenario(typing_scenario_runner: ScenarioRunner) -> Scenario:
    return typing_scenario_runner.scenario


@pytest.fixture
def scenario_runner(typing_scenario_runner: ScenarioRunner) -> ScenarioRunner:
    return typing_scenario_runner



import types

from .. import protocols


class InvalidImportPath(ValueError):
    pass


class ImportPathHelper:
    """
    Helper for creating strings that are valid protocols.ImportPath objects
    """

    def from_cls(self, cls: type) -> protocols.ImportPath:
        """
        Given some class return an import path to it
        """
        return self(f"{cls.__module__}.{cls.__qualname__}")

    def cls_module(self, cls: type) -> protocols.ImportPath:
        """
        Given some cls return an import path to the module that defined it
        """
        return self(cls.__module__)

    def from_module(self, module: types.ModuleType) -> protocols.ImportPath:
        """
        Given some module return an import path to it
        """
        return self(module.__name__)

    def split(
        self, path: protocols.ImportPath
    ) -> tuple[protocols.ImportPath, protocols.ImportPath]:
        """
        Split a path into it's namespace and name.

        So `my.code.Thing` splits into (`my.code`, `Thing`)

        If the path is not namespaced then a InvalidImportPath will be raised
        """
        if "." not in path:
            raise InvalidImportPath(f"Provided path was not namespaced: '{path}'")
        namespace, name = path.rsplit(".", 1)
        return protocols.ImportPath(namespace), protocols.ImportPath(name)

    def __call__(self, path: str) -> protocols.ImportPath:
        """
        Return a string as a protocols.ImportPath type.

        If the string is not a valid import then a InvalidImportPath will be raised
        """
        if not all(part and part.isidentifier() for part in path.split(".")):
            raise InvalidImportPath(f"Provided path was not a valid python import path: '{path}'")
        return protocols.ImportPath(path)


ImportPath = ImportPathHelper()



from __future__ import annotations

import dataclasses
import functools

import pytest
from typing_extensions import Self

from extended_mypy_django_plugin.django_analysis import (
    Field,
    ImportPath,
    Model,
    Module,
    Project,
    protocols,
)


@pytest.fixture(autouse=True)
def _ensure_django_loaded(loaded_django_example: protocols.Loaded[Project]) -> None:
    """
    Make sure the loaded_django_example fixture is active so that we can import our djangoexample models inside tests
    """


@dataclasses.dataclass
class EmptyField:
    @classmethod
    def create(
        cls, *, model_import_path: protocols.ImportPath, field: protocols.DjangoField
    ) -> Self:
        return cls(model_import_path=model_import_path, field=field)

    def __post_init__(self) -> None:
        self.field_type = ImportPath.from_cls(self.field.__class__)

    model_import_path: protocols.ImportPath
    field: protocols.DjangoField

    field_type: protocols.ImportPath = dataclasses.field(init=False)
    related_model: protocols.ImportPath | None = None


class TestModule:
    def test_interpreting_a_module(self) -> None:
        import djangoexample.exampleapp.models

        module_import_path = ImportPath("djangoexample.exampleapp.models")

        module = Module.create(
            model_creator=functools.partial(Model.create, field_creator=EmptyField.create),
            import_path=module_import_path,
            module=(mod := djangoexample.exampleapp.models),
            models=[mod.Parent, mod.Child1],
        )

        defined_models = [
            Model(
                model_name="Parent",
                module_import_path=module_import_path,
                import_path=(mip := ImportPath(f"{module_import_path}.Parent")),
                is_abstract=True,
                default_custom_queryset=None,
                all_fields={
                    "one": EmptyField(
                        model_import_path=mip, field=mod.Parent._meta.get_field("one")
                    ),
                },
                models_in_mro=[],
            ),
            Model(
                model_name="Child1",
                module_import_path=module_import_path,
                import_path=(mip := ImportPath(f"{module_import_path}.Child1")),
                is_abstract=False,
                default_custom_queryset=None,
                all_fields={
                    "id": EmptyField(
                        model_import_path=mip, field=mod.Child1._meta.get_field("id")
                    ),
                    "one": EmptyField(
                        model_import_path=mip, field=mod.Child1._meta.get_field("one")
                    ),
                    "two": EmptyField(
                        model_import_path=mip, field=mod.Child1._meta.get_field("two")
                    ),
                },
                models_in_mro=[
                    ImportPath("djangoexample.exampleapp.models.Parent"),
                ],
            ),
        ]

        assert module == Module(
            import_path=ImportPath("djangoexample.exampleapp.models"),
            defined_models={model.import_path: model for model in defined_models},
        )


class TestModel:
    def test_it_can_interpret_an_abstract_model(self) -> None:
        import djangoexample.exampleapp.models

        mod = djangoexample.exampleapp.models

        expected = Model(
            model_name="Parent",
            module_import_path=(
                module_import_path := ImportPath("djangoexample.exampleapp.models")
            ),
            import_path=(mip := ImportPath(f"{module_import_path}.Parent")),
            is_abstract=True,
            default_custom_queryset=None,
            all_fields={
                "one": EmptyField(model_import_path=mip, field=mod.Parent._meta.get_field("one")),
            },
            models_in_mro=[],
        )

        assert (
            Model.create(
                field_creator=EmptyField.create, model=djangoexample.exampleapp.models.Parent
            )
            == expected
        )

    def test_it_can_find_multiple_parent_models(self) -> None:
        import djangoexample.exampleapp.models

        made = Model.create(
            field_creator=EmptyField.create, model=djangoexample.exampleapp.models.Child3
        )

        assert made.models_in_mro == [
            ImportPath("djangoexample.exampleapp.models.Parent2"),
            ImportPath("djangoexample.exampleapp.models.Parent"),
        ]

    def test_it_can_interpret_a_model_with_a_custom_queryset(self) -> None:
        import djangoexample.exampleapp.models

        mod = djangoexample.exampleapp.models

        expected = Model(
            model_name="Child2",
            module_import_path=(
                module_import_path := ImportPath("djangoexample.exampleapp.models")
            ),
            import_path=(mip := ImportPath(f"{module_import_path}.Child2")),
            is_abstract=False,
            default_custom_queryset=ImportPath("djangoexample.exampleapp.models.Child2QuerySet"),
            all_fields={
                "id": EmptyField(model_import_path=mip, field=mod.Child2._meta.get_field("id")),
                "one": EmptyField(model_import_path=mip, field=mod.Child2._meta.get_field("one")),
                "two": EmptyField(model_import_path=mip, field=mod.Child2._meta.get_field("two")),
                "four": EmptyField(
                    model_import_path=mip, field=mod.Child2._meta.get_field("four")
                ),
                "three": EmptyField(
                    model_import_path=mip, field=mod.Child2._meta.get_field("three")
                ),
            },
            models_in_mro=[ImportPath("djangoexample.exampleapp.models.Parent")],
        )

        assert (
            Model.create(
                field_creator=EmptyField.create, model=djangoexample.exampleapp.models.Child2
            )
            == expected
        )

    def test_it_can_interpret_a_model_without_a_custom_queryset(self) -> None:
        import djangoexample.exampleapp.models

        mod = djangoexample.exampleapp.models

        expected = Model(
            model_name="Child1",
            module_import_path=(
                module_import_path := ImportPath("djangoexample.exampleapp.models")
            ),
            import_path=(mip := ImportPath(f"{module_import_path}.Child1")),
            is_abstract=False,
            default_custom_queryset=None,
            all_fields={
                "id": EmptyField(model_import_path=mip, field=mod.Child1._meta.get_field("id")),
                "one": EmptyField(model_import_path=mip, field=mod.Child1._meta.get_field("one")),
                "two": EmptyField(model_import_path=mip, field=mod.Child1._meta.get_field("two")),
            },
            models_in_mro=[ImportPath("djangoexample.exampleapp.models.Parent")],
        )

        assert (
            Model.create(
                field_creator=EmptyField.create, model=djangoexample.exampleapp.models.Child1
            )
            == expected
        )

    def test_it_can_see_reverse_relationships(self) -> None:
        import djangoexample.relations1.models

        mod = djangoexample.relations1.models

        expected = Model(
            model_name="Concrete1",
            module_import_path=(
                module_import_path := ImportPath("djangoexample.relations1.models")
            ),
            import_path=(mip := ImportPath(f"{module_import_path}.Concrete1")),
            is_abstract=False,
            default_custom_queryset=ImportPath(
                "djangoexample.relations1.models.Concrete1QuerySet"
            ),
            all_fields={
                "id": EmptyField(model_import_path=mip, field=mod.Concrete1._meta.get_field("id")),
                "c2s": EmptyField(
                    model_import_path=mip, field=mod.Concrete1._meta.fields_map["c2s"]
                ),
                "thing": EmptyField(
                    model_import_path=mip, field=mod.Concrete1._meta.fields_map["thing"]
                ),
            },
            models_in_mro=[],
        )

        assert (
            Model.create(
                field_creator=EmptyField.create, model=djangoexample.relations1.models.Concrete1
            )
            == expected
        )

    def test_it_can_see_forward_relationships(self) -> None:
        import djangoexample.relations1.models

        mod = djangoexample.relations1.models

        expected = Model(
            model_name="Concrete2",
            module_import_path=(
                module_import_path := ImportPath("djangoexample.relations1.models")
            ),
            import_path=(mip := ImportPath(f"{module_import_path}.Concrete2")),
            is_abstract=False,
            default_custom_queryset=None,
            all_fields={
                "id": EmptyField(model_import_path=mip, field=mod.Concrete2._meta.get_field("id")),
                "concrete1": EmptyField(
                    model_import_path=mip, field=mod.Concrete2._meta.get_field("concrete1")
                ),
                "children": EmptyField(
                    model_import_path=mip, field=mod.Concrete2._meta.get_field("children")
                ),
                "Concrete2_children+": EmptyField(
                    model_import_path=mip,
                    field=mod.Concrete2._meta.fields_map["Concrete2_children+"],
                ),
            },
            models_in_mro=[],
        )

        assert (
            Model.create(
                field_creator=EmptyField.create, model=djangoexample.relations1.models.Concrete2
            )
            == expected
        )


class TestField:
    def test_it_can_interpret_a_plain_field(self) -> None:
        import djangoexample.exampleapp.models

        mod = djangoexample.exampleapp.models

        field = Field.create(
            model_import_path=(mip := ImportPath("djangoexample.exampleapp.models.Child2")),
            field=mod.Child2._meta.get_field("one"),
        )

        assert field == Field(
            model_import_path=mip,
            field_type=ImportPath("django.db.models.fields.CharField"),
            related_model=None,
        )

    def test_it_can_see_reverse_related_fields(self) -> None:
        import djangoexample.relations1.models

        mod = djangoexample.relations1.models

        field = Field.create(
            model_import_path=(mip := ImportPath("djangoexample.relations1.models.Concrete1")),
            field=mod.Concrete1._meta.fields_map["thing"],
        )

        assert field == Field(
            model_import_path=mip,
            field_type=ImportPath("django.db.models.fields.reverse_related.OneToOneRel"),
            related_model=ImportPath("djangoexample.relations2.models.Thing"),
        )

    def test_it_can_see_many_to_many_relations(self) -> None:
        import djangoexample.relations1.models

        mod = djangoexample.relations1.models

        field = Field.create(
            model_import_path=(mip := ImportPath("djangoexample.relations1.models.Concrete2")),
            field=mod.Concrete2._meta.get_field("children"),
        )

        assert field == Field(
            model_import_path=mip,
            field_type=ImportPath("django.db.models.fields.related.ManyToManyField"),
            related_model=ImportPath("djangoexample.relations1.models.Child1"),
        )

    def test_it_can_see_many_to_many_relations_hidden_field(self) -> None:
        import djangoexample.relations1.models

        mod = djangoexample.relations1.models

        field = Field.create(
            model_import_path=(mip := ImportPath("djangoexample.relations1.models.Concrete2")),
            field=mod.Concrete2._meta.fields_map["Concrete2_children+"],
        )

        assert field == Field(
            model_import_path=mip,
            field_type=ImportPath("django.db.models.fields.reverse_related.ManyToOneRel"),
            # Empty because it refers to a many to many model class that doesn't exist
            related_model=None,
        )

    def test_it_can_see_forieng_key(self) -> None:
        import djangoexample.relations1.models

        mod = djangoexample.relations1.models

        field = Field.create(
            model_import_path=(mip := ImportPath("djangoexample.relations1.models.Concrete2")),
            field=mod.Concrete2._meta.get_field("concrete1"),
        )

        assert field == Field(
            model_import_path=mip,
            field_type=ImportPath("django.db.models.fields.related.ForeignKey"),
            related_model=ImportPath("djangoexample.relations1.models.Concrete1"),
        )



import dataclasses
from typing import TYPE_CHECKING, Generic, cast

from .. import protocols


@dataclasses.dataclass(frozen=True, kw_only=True)
class ConcreteModelsDiscovery(Generic[protocols.T_Project]):
    """
    The default implementation that determines the concrete models for each model.

    For abstract models, this is all the concrete children.

    For concrete models, this is the model by itself.
    """

    def __call__(
        self,
        loaded_project: protocols.Loaded[protocols.T_Project],
        all_models: protocols.ModelMap,
        /,
    ) -> protocols.ConcreteModelsMap:
        result: dict[protocols.ImportPath, list[protocols.ImportPath]] = {}

        for import_path, model in all_models.items():
            if model.is_abstract:
                if import_path not in result:
                    result[import_path] = []
                continue

            result[import_path] = [model.import_path]
            for mro_import_path in model.models_in_mro:
                found = all_models[mro_import_path]
                if found.is_abstract:
                    if mro_import_path not in result:
                        result[mro_import_path] = []
                    result[mro_import_path].append(import_path)

        return {
            import_path: [all_models[path] for path in concrete]
            for import_path, concrete in result.items()
        }


if TYPE_CHECKING:
    _STA: protocols.P_ConcreteModelsDiscovery = cast(
        ConcreteModelsDiscovery[protocols.P_Project], None
    )
